module Tree where

import Control.Lens.Getter ((^.))
import Data.List.Split (chunksOf)
import Data.Map.Strict qualified as Map
import Data.Map.Strict (insert)
import Data.Map.Strict.Internal (merge, preserveMissing, zipWithMatched)
import Data.Text qualified as T
import Data.Text.Format (format)
import Data.Text.Lazy qualified as TL
import GHC.Generics
import HAL
import HAL.Client
import HAL.Doc.Struct
import Protolude
--import Text.Printf

formatParentIdRequest :: [Struct] -> Maybe Text
formatParentIdRequest [] = Nothing
formatParentIdRequest (x:[]) = Just . T.pack . show $ _struct_docid x
formatParentIdRequest (x:xs) =
  (Just . T.pack . show $ _struct_docid x)
  <> (Just " || ")
  <> formatParentIdRequest xs


ds2Child :: [Struct] -> IO [Struct]
ds2Child ds =  do
  rs <- sequence $ runStructureRequest <$> formatedRequests ds
  return . concat $ (^. docs) <$> rights rs
  where formatedRequest docs' = Just "parentDocid_i:(" <> formatParentIdRequest docs' <> Just ")"
        formatedRequests docs'' = formatedRequest <$> chunksOf 100 docs''

fetchChildren :: [Struct] -> IO [Struct]
fetchChildren [] = pure []
fetchChildren ds = (ds <>) <$> (fetchChildren =<< ds2Child ds)

isChildOf :: Struct -> Struct -> Bool
isChildOf (Struct i _ _) (Struct _ _ p') =
  not . null $ filter (\id -> id == (T.pack $ show i)) p'

data DocTree = DocTree Struct Int [DocTree]
  deriving (Show, Generic)

buildTree :: Int -> [Struct] -> Struct -> DocTree
buildTree depth docs' id = DocTree id depth (buildTree (depth + 1) docs' <$> children)
  where
    children = filter (isChildOf id) docs'

formatTree :: DocTree -> Text
formatTree (DocTree doc depth children) =
  TL.toStrict $ format "{}{}\n{}" [ addSpace
                                  , show doc
                                  , T.concat $ formatTree <$> children ]
  where
    addSpace :: Text
    addSpace
          | depth > 0 = "├" <> (T.replicate (depth * depth + depth) "─")
          | otherwise = "🌲"

findDeepest :: Map Int Int -> DocTree -> Map Int Int
findDeepest m (DocTree (Struct docid _ _) depth children) =
  mergeMap map1 maps
  where map1 = insert docid depth m
        maps = foldl mergeMap Map.empty $ findDeepest m <$> children
        mergeMap m1 m2 = merge preserveMissing preserveMissing (zipWithMatched whenMatch) m1 m2
        whenMatch _k v v' = if v > v' then v else v'

isDeep :: Map Int Int -> DocTree -> Bool
isDeep m (DocTree (Struct id _ _) depth _) = depth >= (fromMaybe 0 $ Map.lookup id m)

removeDuplicate :: Map Int Int -> DocTree -> DocTree
removeDuplicate deepMap (DocTree doc depth children) =
  DocTree doc depth (removeDuplicate deepMap <$> filter (isDeep deepMap) children)

noDuplicateTree :: DocTree -> DocTree
noDuplicateTree tree = removeDuplicate (findDeepest Map.empty tree) tree
